from collections import namedtuple
import logging
from typing import Any, Dict, Generator
from jf_ingest import diagnostics, logging_helper
from jira import JIRA

# jira renamed this between api versions for some reason
try:
    from jira.resources import AgileResource as AGILE_BASE_REST_PATH
except ImportError:
    from jira.resources import GreenHopperResource as AGILE_BASE_REST_PATH

logger = logging.getLogger(__name__)


def get_jira_connection(config, creds, max_retries=3) -> JIRA:
    kwargs = {
        "server": config.jira_url,
        "max_retries": max_retries,
        "options": {
            "agile_rest_path": AGILE_BASE_REST_PATH,
            "verify": not config.skip_ssl_verification,
        },
    }

    if creds.jira_username and creds.jira_password:
        kwargs["basic_auth"] = (creds.jira_username, creds.jira_password)
    elif creds.jira_bearer_token:
        kwargs["options"]["headers"] = {
            "Authorization": f"Bearer {creds.jira_bearer_token}",
            "Cache-Control": "no-cache",
            "Content-Type": "application/json",
            "Accept": "application/json",
            "X-Atlassian-Token": "no-check",
        }
    else:
        raise RuntimeError(
            "No valid Jira credentials found! Check your JIRA_USERNAME, JIRA_PASSWORD, or JIRA_BEARER_TOKEN environment variables."
        )

    jira_connection = JIRA(**kwargs)

    jira_connection._session.headers[
        "User-Agent"
    ] = f'jellyfish/1.0 ({jira_connection._session.headers["User-Agent"]})'

    return jira_connection


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_fields(
    jira_connection: JIRA, include_fields=[], exclude_fields=[]
) -> list[dict]:
    logger.info("downloading jira fields... ")

    filters = []
    if include_fields:
        filters.append(lambda field: field["id"] in include_fields)
    if exclude_fields:
        filters.append(lambda field: field["id"] not in exclude_fields)

    fields = [
        field
        for field in jira_connection.fields()
        if all(filter(field) for filter in filters)
    ]

    logger.info("✓")
    return fields


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_projects_and_versions(
    jira_connection: JIRA,
    include_projects,
    exclude_projects,
    include_categories,
    exclude_categories,
) -> list[dict]:
    return []


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_users(
    jira_connection,
    gdpr_active,
    quiet=False,
    required_email_domains=None,
    is_email_required=False,
) -> list[dict]:
    return []


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_resolutions(jira_connection) -> list[dict]:
    return []


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_issuetypes(jira_connection, project_ids) -> list[dict]:
    return []


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_issuelinktypes(jira_connection) -> list[dict]:
    return []


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_priorities(jira_connection) -> list[dict]:
    return []


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_projects_and_versions(
    jira_connection,
    include_projects,
    exclude_projects,
    include_categories,
    exclude_categories,
) -> list[dict]:
    return []


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_boards_and_sprints(
    jira_connection: JIRA, project_ids, download_sprints
) -> list[dict]:
    return []


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def get_issues(jira_connection, issue_jql, start_at, batch_size) -> list[dict]:
    return []


# TODO: Make this a dataclass. Not a fan of namedtuple
IssueMetadata = namedtuple("IssueMetadata", ("key", "updated"))


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_all_issue_metadata(
    jira_connection,
    all_project_ids,
    earliest_issue_dt,
    num_parallel_threads,
    issue_filter,
) -> Dict[int, IssueMetadata]:
    return []


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def detect_issues_needing_sync(
    issue_metadata_from_jira: Dict[int, IssueMetadata],
    issue_metadata_from_jellyfish: Dict[int, IssueMetadata],
) -> list[dict]:
    return []


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_worklogs(jira_connection, issue_ids, work_logs_pull_from) -> list[dict]:
    return []


# Returns an array of CustomFieldOption items
@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_customfieldoptions(jira_connection, project_ids) -> list[dict]:
    return []


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_statuses(jira_connection) -> list[dict]:
    return []


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def detect_issues_needing_re_download(
    downloaded_issue_id_and_key_tuples: set[tuple[str, str]],
    issue_metadata_from_jellyfish,
    issue_metadata_addl_from_jellyfish,
) -> list[dict]:
    return []


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_necessary_issues(
    jira_connection,
    issue_ids_to_download,
    include_fields,
    exclude_fields,
    num_parallel_threads,
    suggested_batch_size: int = 2000,
) -> Generator[Any, None, None]:
    return []



import os

import requests_mock
from unittest import TestCase

from jf_ingest.jf_jira.auth import JiraAuthConfig, get_jira_connection
from jf_ingest.jf_jira.downloaders import download_fields

MOCK_SERVER_INFO_RESP = (
    '{"baseUrl":"https://test-co.atlassian.net","version":"1001.0.0-SNAPSHOT",'
    '"versionNumbers":[1001,0,0],"deploymentType":"Cloud","buildNumber":100218,'
    '"buildDate":"2023-03-16T08:21:48.000-0400","serverTime":"2023-03-17T16:32:45.255-0400",'
    '"scmInfo":"9999999999999999999999999999999999999999","serverTitle":"JIRA",'
    '"defaultLocale":{"locale":"en_US"}} '
)


def get_connection(mocker: requests_mock.Mocker):

    auth_config = JiraAuthConfig(
        url="https://test-co.atlassian.net/",
        personal_access_token="asdf",
        company_slug="test_co",
    )
    # you can test behavior against a live jira instance by setting an email as `jira_username`
    # and storing a generated token in env and retrieving like so:
    # creds.jira_bearer_token = os.environ.get('JIRA_TOKEN')

    # https://test-co.atlassian.net/rest/api/2/serverInfo
    mocker.register_uri(
        "GET",
        "https://test-co.atlassian.net/rest/api/2/serverInfo",
        text=f"{MOCK_SERVER_INFO_RESP}",
    )
    jira_conn = get_jira_connection(config=auth_config, max_retries=1)

    return jira_conn


class TestJiraDownload(TestCase):
    URL_BASE = "https://test-co.atlassian.net/rest/api/2/search"

    start_at = 0
    max_results = 100

    jira_connection = None
    mock_field_response = None

    @classmethod
    def setUpClass(cls):
        with open(
            f"{os.path.dirname(__file__)}/fixtures/jira_fields_response.json", "r"
        ) as file:
            cls.mock_field_response = file.read()
        with requests_mock.Mocker() as m:
            m.register_uri(
                "GET",
                "https://test-co.atlassian.net/rest/api/2/field",
                text=f"{cls.mock_field_response}",
            )
            cls.jira_connection = get_connection(m)

    def test_get_fields_once(self):
        expected_field = {
            "id": "statuscategorychangedate",
            "key": "statuscategorychangedate",
            "name": "Status Category Changed",
            "custom": False,
            "orderable": False,
            "navigable": True,
            "searchable": True,
            "clauseNames": ["statusCategoryChangedDate"],
            "schema": {"type": "datetime", "system": "statuscategorychangedate"},
        }

        with requests_mock.Mocker() as m:
            m.register_uri(
                "GET",
                "https://test-co.atlassian.net/rest/api/2/field",
                text=self.mock_field_response,
            )
            fields = download_fields(
                jira_connection=self.jira_connection,
                include_fields=[],
                exclude_fields=[],
            )

        self.assertEqual(4, len(fields))
        self.assertEqual(expected_field, fields[0])
        for field in fields:
            self.assertIn("id", field.keys())
            self.assertIn("key", field.keys())
            self.assertIn("name", field.keys())

    def test_get_fields_exclude_list(self):
        with requests_mock.Mocker() as m:
            m.register_uri(
                "GET",
                "https://test-co.atlassian.net/rest/api/2/field",
                text=self.mock_field_response,
            )
            fields = download_fields(
                jira_connection=self.jira_connection,
                include_fields=[],
                exclude_fields=["customfield_10070"],
            )

        self.assertEqual(3, len(fields))

    def test_get_fields_include_list(self):
        with requests_mock.Mocker() as m:
            m.register_uri(
                "GET",
                "https://test-co.atlassian.net/rest/api/2/field",
                text=self.mock_field_response,
            )
            fields = download_fields(
                jira_connection=self.jira_connection,
                include_fields=["customfield_10070"],
                exclude_fields=[],
            )

        self.assertEqual(1, len(fields))
